# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: MIT-0

import awsiot.greengrasscoreipc
import awsiot.greengrasscoreipc.client as client
from awsiot.greengrasscoreipc.model import QOS, PublishToIoTCoreRequest
import cv2
import inference
import io
import json
import logging
from PIL import Image, ImageDraw, ImageFont
import os
import time
import sys

current_milli_time = lambda: int(round(time.time() * 1000))

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger()

STREAM_IMAGE_FILE = os.environ.get("STREAM_IMAGE_FILE", "/home/nvidia/detected.jpg")
MODEL_PATH = os.environ.get("MODEL_PATH", "/home/nvidia/takenoko")
CUDA_BIN_PATH = os.environ.get("CUDA_BIN_PATH", "/usr/local/cuda/bin")
os.environ["PATH"] = "{};{}".format(CUDA_BIN_PATH, os.environ["PATH"])
MXNET_HOME_PATH = os.environ.get("MXNET_HOME_PATH", "/home/nvidia/mxnet")
sys.path.append(MXNET_HOME_PATH + "/python")
TOPIC = "bf202205/action"
TEXTSIZE = 14
LINEWIDTH = 4
VIDEO_WIDTH = 960
VIDEO_HEIGHT = 720

ipc_client = awsiot.greengrasscoreipc.connect()

cap = cv2.VideoCapture(0)  # video device 0
cap.set(cv2.CAP_PROP_FRAME_WIDTH, VIDEO_WIDTH)
cap.set(cv2.CAP_PROP_FRAME_HEIGHT, VIDEO_HEIGHT)

width = cap.get(cv2.CAP_PROP_FRAME_WIDTH)
height = cap.get(cv2.CAP_PROP_FRAME_HEIGHT)
fps = cap.get(cv2.CAP_PROP_FPS)
logger.info("width:{} height:{} fps:{}".format(width, height, fps))
logger.error(os.environ.get)

# モデルのロード
logger.info("load model")
model = inference.model_fn(MODEL_PATH)


def publish(inference_class, predict_duration):
    message = {"type": inference_class, "predict_duration": predict_duration}
    operation = ipc_client.new_publish_to_iot_core()
    operation.activate(
        PublishToIoTCoreRequest(
            topic_name=TOPIC, qos=QOS.AT_LEAST_ONCE, payload=json.dumps(message)
        )
    )
    logger.info("publish message:" + json.dumps(message))


font = ImageFont.truetype(
    "/usr/share/fonts/truetype/dejavu/DejaVuSerif.ttf", size=TEXTSIZE
)

dir_path = os.path.dirname(STREAM_IMAGE_FILE)
if not os.path.exists(dir_path):
    logger.info("create stream image path")
    os.makedirs(dir_path)

last_inference_class = ""
while True:
    ret, image_data = cap.read()
    if ret:
        logger.debug("image captured")
    else:
        logger.error("image capture faild")
        time.sleep(5)
        continue

    frame = cv2.cvtColor(image_data, cv2.COLOR_BGR2RGB)
    org_img = Image.fromarray(frame)
    jpg = io.BytesIO()
    org_img.save(jpg, format="jpeg")
    img_bin = jpg.getvalue()
    jpg.close()

    logger.debug("start predict")
    t1 = current_milli_time()
    pred = json.loads(
        inference.transform_fn(
            model, img_bin, "application/x-image", "application/json"
        )
    )
    predict_duration = current_milli_time() - t1
    logger.debug(f"predict duration: {predict_duration} ms")

    # 推論結果の展開
    bboxes, classes, confidences = (
        pred["normalized_boxes"],
        pred["classes"],
        pred["scores"],
    )
    # 推論結果の描画
    draw = ImageDraw.Draw(org_img)
    detected = False
    inference_class = None
    for i in range(len(bboxes)):
        # 信頼度スコアが0.8以上のみ描画する
        if confidences[i] > 0.8:
            detected = True
            # 矩形の相対座標を取得
            left, top, right, bottom = bboxes[i]
            # 矩形の相対座標を絶対座標に変換
            left = org_img.size[0] * left
            top = org_img.size[1] * top
            right = org_img.size[0] * right
            bottom = org_img.size[1] * bottom
            # 検出した物体の ID を take/kino に読み替える
            if int(classes[i]) == 0:
                inference_class = "kinoko"
                color = "red"
            else:
                inference_class = "takenoko"
                color = "blue"
            # 矩形を描画する
            draw.rectangle(
                [(left, top), (right, bottom)], outline=color, width=LINEWIDTH
            )
            # 矩形の左上に描画する信頼度スコアの取得
            title = f"{inference_class} {str(round(confidences[i],3))}"
            # テキストを描画する場所を取得
            txpos = (left, top - TEXTSIZE - LINEWIDTH // 2)
            # 描画するテキストのサイズを取得
            txw, txh = draw.textsize(title, font=font)
            # テキストの背景用の矩形を描画
            draw.rectangle(
                [txpos, (left + txw, top)], outline=color, fill=color, width=LINEWIDTH
            )
            # テキストを描画
            draw.text(txpos, title, fill="white", font=font)
            # 描画は一つだけ
            break

    # 表示用に画像をファイルに書き出す
    org_img.save(STREAM_IMAGE_FILE)
    if detected:
        if not last_inference_class == inference_class:
            # AWS IoT Coreにpublish
            publish(inference_class, predict_duration)
            last_inference_class = inference_class
