# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: MIT-0

import json
import logging
import os
import re
from typing import Any
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import gluoncv
import mxnet as mx
import numpy as np
from constants import constants
import encoder


logging.basicConfig(format="%(asctime)s %(message)s", level=logging.INFO)
CTX = mx.gpu() if mx.context.num_gpus() > 0 else mx.cpu()


def _get_model_name(model_dir: str) -> str:
    """Extract the model name from model_info file in model_dir."""
    try:
        with open(os.path.join(model_dir, constants.MODEL_INFO_FILE_NAME)) as f:
            model_info = json.loads(f.read())
            model_name = model_info[constants.MODEL_NAME]
            return model_name
    except Exception as e:
        logging.info(
            f"Could not extract {constants.MODEL_NAME} from {constants.MODEL_INFO_FILE_NAME}. Error:{str(e)}"
        )
        raise


def _is_finetuned(model_dir: str) -> bool:
    """Read model_info file and check if the model is finetuned or pre-trained."""
    try:
        with open(os.path.join(model_dir, constants.MODEL_INFO_FILE_NAME)) as f:
            model_info = json.loads(f.read())
            if "finetuned" in model_info:
                return model_info["finetuned"]
            else:
                return False
    except Exception as e:
        logging.info(f"Could not read {constants.MODEL_INFO_FILE_NAME}. Error:{str(e)}")
        raise


def _get_file(directory: str, regex: str) -> Optional[str]:
    """Explore the directory tree from the argument provided and return the first filename matching the regex.

    Args:
        directory: directory containing the file.
        regex: regex pattern of the file to be found.

    Returns:
        file name matching regex. If no such file, return None
    """
    for _, _, files in os.walk(directory):
        for file in files:
            matches = re.compile(regex).match(file)
            if matches:
                return matches.group()
    return None


class MXNetOD:
    """Instantiate with a model.

    It takes in an image and returns boxes, labels and scores for each predicted object encoded in
    json/csv/numpy.array format.
    """

    def __init__(self, model, model_name: str, labels: List[Any]):
        """Save the model and list of labels."""
        self.model = model
        self.model_name = model_name
        self.labels = labels

    def forward(self, input_data):
        """Make one forward pass of the model and return the output.

        Args:
            input_data: Input image with rgb channels as a torch tensor.

        Returns:
            output of the network on the input image. A list of single models.common.Detections object.
        """
        return self.model(input_data)

    @classmethod
    def decode_results(cls, output):
        """Return bounding boxes, class IDs and scores as a list."""
        class_IDs = [x[0] for x in output[0][0].asnumpy()]
        scores = [x[0] for x in output[1][0].asnumpy()]
        bounding_boxes = output[2][0].asnumpy()
        valid_boxes_indices = [
            i for i, class_id in enumerate(class_IDs) if class_id >= 0
        ]
        class_IDs = [class_IDs[i] for i in valid_boxes_indices]
        scores = [scores[i] for i in valid_boxes_indices]
        bounding_boxes = [bounding_boxes[i] for i in valid_boxes_indices]
        return [bounding_boxes, class_IDs, scores]

    @classmethod
    def normalize_predictions(
        cls, predictions: list, width: int, height: int
    ) -> [np.array, np.array, np.array]:
        """Rescale the bounding boxes to have co-ordinates between 0 and 1."""
        boxes, labels, scores = predictions

        boxes = np.array(boxes)
        scaling_factor = np.array([width, height, width, height])
        return [boxes / scaling_factor, labels, scores]

    @classmethod
    def encode(cls, predictions, accept):
        """Encode results with accept."""
        return encoder.encode(predictions, accept)

    def get_top_k_predictions(
        self, predictions: [np.array, np.array, np.array], k: int
    ) -> Tuple[List[np.array], List[int], List[float]]:
        """Return the top k predictions with the highest scores."""

        boxes, labels, scores = predictions
        k = min(k, boxes.shape[0])
        top_k_indices = np.argpartition(scores, -k)[-k:]
        labels_top_k = [labels[int(idx)] for idx in top_k_indices]
        scores_top_k = [scores[int(idx)] for idx in top_k_indices]
        boxes_top_k = [boxes[int(idx)] for idx in top_k_indices]

        return boxes_top_k, labels_top_k, scores_top_k

    def __call__(
        self,
        input_data,
        content_type="application/json",
        accept="application/json",
        **kwargs,
    ):
        """Invoke the model, decode the prediction and return encoded results."""
        _, _, height, width = input_data.shape
        model_output = self.forward(input_data)
        predictions = self.decode_results(model_output)
        normalized_predictions = self.normalize_predictions(
            predictions=predictions, width=width, height=height
        )

        # If only a subset of predictions are requested, we return the top k predictions as per the scores.
        n_predictions_search = re.search(
            f"{constants.N_PREDICTIONS_KEY}=[0-9]*", accept
        )
        if n_predictions_search is not None:
            n_prediction_pattern = n_predictions_search.group(0)
            n_output_predictions = int(
                n_prediction_pattern.lstrip(f"{constants.N_PREDICTIONS_KEY}=")
            )
            normalized_predictions = self.get_top_k_predictions(
                normalized_predictions, k=n_output_predictions
            )
            accept = accept.replace(n_prediction_pattern, "")

        output = {
            constants.NORMALIZED_BOXES: normalized_predictions[0],
            constants.CLASSES: normalized_predictions[1],
            constants.SCORES: normalized_predictions[2],
        }

        # If the response_content_type is verbose, we return the labels as well as the original model output.
        if constants.VERBOSE_EXTENSION in accept:
            output[constants.LABELS] = self.labels
            output[constants.MX_MODEL_OUTPUT] = [x.asnumpy() for x in model_output]
            accept = accept.replace(constants.VERBOSE_EXTENSION, "")

        return self.encode(output, accept=accept)


def _get_labels(model_dir: str) -> List[Any]:
    """Read, parse and return the labels stored as JSON in model directory."""

    with open(os.path.join(model_dir, constants.LABELS_INFO_FILE_NAME), "r") as txtFile:
        labels = json.loads(txtFile.read())[constants.LABELS]
    return labels


def model_fn(model_dir: str) -> MXNetOD:
    """Create our inference task as a delegate to the model.

    This runs only once per one worker.
    """

    model_name = _get_model_name(model_dir=model_dir)
    is_finetuned = _is_finetuned(model_dir=model_dir)
    labels = _get_labels(model_dir=model_dir)
    try:
        model = gluoncv.model_zoo.get_model(
            model_name, pretrained=True, root=model_dir, ctx=CTX
        )
        if is_finetuned:
            classes = range(len(labels))
            model.reset_class(classes)
            model.load_parameters(os.path.join(model_dir, "finetuned_model.params"))
        model.hybridize()
        return MXNetOD(model=model, model_name=model_name, labels=labels)
    except Exception:
        logging.exception("Failed to load model")
        raise


def transform_fn(
    task: MXNetOD, input_data: bytearray, content_type: str, accept: str
) -> Union[str, bytes]:
    """Make predictions against the model and return a serialized response.

    The function signature conforms to the SageMaker contract.

    Args:
        task: model loaded by model_fn, in our case is one of the Task.
        input_data: the request data.
        content_type: the request content type.
        accept: accept header expected by the client.

    Returns:
        the serialized prediction result.
    """
    if content_type == constants.REQUEST_CONTENT_TYPE:
        try:
            input_data = mx.image.imdecode(bytes(input_data))
            if task.model_name in constants.ssd_models:
                transform = gluoncv.data.transforms.presets.ssd.transform_test
            elif task.model_name in constants.center_net_models:
                transform = gluoncv.data.transforms.presets.center_net.transform_test
            elif task.model_name in constants.faster_rcnn_models:
                transform = gluoncv.data.transforms.presets.rcnn.transform_test
            elif task.model_name in constants.yolo_models:
                transform = gluoncv.data.transforms.presets.yolo.transform_test
            else:
                logging.info(f"Model name {task.model_name} not supported")
                raise
            input_data, _ = transform(input_data, short=512)
            input_data = input_data.as_in_context(CTX)
        except Exception:
            logging.exception("Failed to read image")
            raise
        try:
            output = task(
                input_data=input_data, content_type=content_type, accept=accept
            )
            return output
        except Exception:
            logging.exception("Failed to do inference")
            raise
    raise ValueError(
        '{{"error": "unsupported content type {}"}}'.format(content_type or "unknown")
    )
